from enum import Enum

import typer

from changelog.cli.state import get_changelog, save_changelog

app = typer.Typer()


@app.callback()
def config():
    """Manage config encoded in changelog."""


class ConfigField(Enum):
    release_link_format = "release_link_format"
    breaking_change_token = "breaking_change_token"


@app.command()
def get(
    field: ConfigField = typer.Option(..., help="Config field to retrieve."),
):
    """Retrieve a config value from the changelog."""
    changelog = get_changelog()
    typer.echo(changelog.config.get(field.name, "not set"))


@app.command()
def set(
    field: ConfigField = typer.Option(..., help="Config field to set."),
    value: str = typer.Option(..., help="Value to set field to."),
):
    """Set a config value in the changelog."""
    changelog = get_changelog()
    changelog.config.set(field.name, value)
    save_changelog(changelog)



from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import Optional

from changelog.exceptions import ChangelogParseError
from changelog.model import Changelog, Entry, ReleaseSection, ReleaseTag


@dataclass
class ParserState:
    changelog: Changelog = field(default_factory=Changelog)
    release_tag: Optional[tuple[ReleaseTag, Optional[str]]] = None
    change_type: Optional[str] = None
    entry_stack: list[tuple[Entry, int]] = field(default_factory=list)

    @property
    def root_entry(self) -> Optional[Entry]:
        if not self.entry_stack:
            return None
        return self.entry_stack[0][0]

    def parent_entry(self, indentation_chars: int) -> Optional[Entry]:
        if not self.entry_stack or indentation_chars == self.entry_stack[0][1]:
            # Top level entry
            return None
        if indentation_chars > self.entry_stack[-1][1]:
            # If its more indented than the previous entry, its a child of that entry
            return self.entry_stack[-1][0]
        # Remove more nested items from the stack
        while indentation_chars < self.entry_stack[-1][1]:
            self.entry_stack.pop(-1)
        if indentation_chars != self.entry_stack[-1][1]:
            raise ChangelogParseError("Bad indentation")
        self.entry_stack.pop(-1)
        return self.entry_stack[-1][0]

    def flush(self) -> None:
        if self.release_tag and self.change_type and self.root_entry:
            tag, timestamp = self.release_tag
            self.changelog.releases.setdefault(tag, ReleaseSection(entries={}, timestamp=timestamp)).entries.setdefault(
                self.change_type, []
            ).append(self.root_entry)
            self.entry_stack = []


def loads(text: str, tab_indent: int = 2) -> Changelog:
    parser_state = ParserState()
    text = text.replace("\t", tab_indent * " ")
    for index, line in enumerate(text.splitlines()):
        if (release_header_match := re.match(r"^## \[(?P<tag>.+)\]( +- +(?P<date>\d+\-\d+\-\d+))?", line)) :
            # New tags are level-two headings, and must be linked.
            # They optionally include a timestamp.
            parser_state.flush()
            match_dict = release_header_match.groupdict()
            tag = ReleaseTag(match_dict["tag"])
            timestamp = match_dict.get("date")
            parser_state.release_tag = tag, timestamp
            parser_state.changelog.releases.setdefault(tag, ReleaseSection(entries={}, timestamp=timestamp))
            continue
        if not parser_state.release_tag:
            # If release_tag is not set, assume we are parsing header text
            parser_state.changelog.header += f"\n{line}"
            continue
        if (
            change_type_match := re.match(
                r"^### (?P<change_type>Security|Deprecated|Added|Changed|Removed|Fixed)$", line
            )
        ) :
            # Change types are grouped under level 3 headings.
            parser_state.flush()
            parser_state.change_type = change_type_match.groupdict()["change_type"]
            continue
        if (entry_start_match := re.match(r"^ *(\*|\+|-) (?P<sub_entry_start>.+)", line)) :
            # New entry start
            indentation_chars = len(line) - len(line.lstrip())
            entry = Entry(text=entry_start_match.groupdict()["sub_entry_start"])
            try:
                parent_entry = parser_state.parent_entry(indentation_chars)
            except ChangelogParseError:
                raise ChangelogParseError(f"Bad indentation at line {index}: {line!r}")
            if not parent_entry:
                # Must be top-level
                parser_state.flush()
            else:
                # New sub entry
                parent_entry.children.append(entry)
            parser_state.entry_stack.append((entry, indentation_chars))
            continue
        if parser_state.entry_stack and (entry_continued_match := re.match(r"^ *(?P<entry_continued>.+)", line)):
            # Multi-line continuation of entry text.
            indentation_chars = len(line) - len(line.lstrip())
            if indentation_chars < parser_state.entry_stack[-1][1] + 2:
                raise ChangelogParseError(f"Line {index} is not indented enough to be a continuation: {line!r}")
            parser_state.entry_stack[-1][0].text += " " + entry_continued_match.groupdict()["entry_continued"].lstrip()
            continue
        if (link_match := re.match(r"^\[(?P<link_name>.+)\]: (?P<link_target>.+)$", line)) :
            # Links follow the format [{link_name}]: http://example.com/link/target
            parser_state.flush()
            match_dict = link_match.groupdict()
            link_name = match_dict["link_name"]
            link_target = match_dict["link_target"]
            if link_name.startswith("_") and link_name[1:] in parser_state.changelog.config.fields:
                # Check if the link is actually a config field in disguise
                field_name = link_name[1:]
                config = parser_state.changelog.config
                field_parser = config.fields[field_name].metadata.get("parse", lambda _: _)
                setattr(config, link_name[1:], field_parser(link_target))
                continue
            parser_state.changelog.links[link_name] = link_target
            continue
        if not line.strip() or "nothing here" in line.lower():
            # Blank lines terminate the previous entry, but are otherwise ignored.
            parser_state.flush()
            continue
        raise ChangelogParseError(f"Invalid changelog at line {index}: {line!r}")
    parser_state.flush()
    parser_state.changelog.header = parser_state.changelog.header.lstrip()
    parser_state.changelog.validate()
    return parser_state.changelog


def load_from_file(path: str = "CHANGELOG.md") -> Changelog:
    with open(path, "r") as file:
        contents = file.read()
    return loads(contents)



import re
from string import Formatter
from typing import TypeVar, Union, overload

_NOT_PASSED = object()


DefaultT = TypeVar("DefaultT")


@overload
def reverse_format(string: str, format_spec: str) -> dict[str, str]:
    ...  # pragma: no cover


@overload
def reverse_format(string: str, format_spec: str, default: DefaultT) -> Union[dict[str, str], DefaultT]:
    ...  # pragma: no cover


def reverse_format(string, format_spec, default=_NOT_PASSED, multiline: bool = False):
    """Inverse of `format`.

    Returns a mapping of the format_spec field names to their values in `string`.

    :param string: The string to extract values from.
    :param format_spec: The format spec which describes the string fields. Field names must
        be valid Python identifiers.
    :param default: Optional value to return if string does not match the format spec.
    :param multiline: If true, allow parameters values to include newlines.
    :raises ValueError: if format spec is invalid, or does not match the string and no
        default is provided.
    """
    pattern = format_spec_to_regex(format_spec, multiline=multiline)
    if not (match := pattern.match(string)):
        if default is _NOT_PASSED:
            raise ValueError(f"String {string!r} does not match format {format_spec!r}")
        return default
    return match.groupdict()


def format_spec_to_regex(format_spec: str, multiline: bool = False) -> re.Pattern:
    """Convert a Python format spec string to a regex pattern with named groups.

    :param format_spec: The format spec to convert.
    :param multiline: If true, allow parameters to include newlines. Adds the re.DOTALL flag
        under-the-hood.
    """
    output = re.escape(format_spec)
    group_name_pattern = re.compile(r"\\{([a-zA-Z_][a-zA-Z0-9_]*)\\}")
    group_names = set(group_name_pattern.findall(output))
    if (invalid_names := _format_spec_field_names(format_spec) - group_names) :
        raise ValueError(
            f"Unsupported field names in format string, only valid Python identifiers are supported: {invalid_names}"
        )
    for group_name in group_names:
        # Set the occurence to declare the group
        output = re.sub(r"\\{(" + group_name + r")\\}", r"(?P<\1>.*)", output, count=1)
        # Set remaining occurences to backreference the group
        output = re.sub(r"\\{(" + group_name + r")\\}", r"(?P=\1)", output)
    flags = [re.DOTALL] if multiline else []
    return re.compile(output, *flags)


def _format_spec_field_names(format_spec: str) -> set[str]:
    """Extract field names from a format spec."""
    return {name for _, name, __, ___ in Formatter().parse(format_spec) if name is not None}



from __future__ import annotations

import re
from collections import OrderedDict
from dataclasses import Field, dataclass, field
from datetime import date
from enum import Enum
from typing import Literal, Optional, cast
from urllib.parse import quote_plus, unquote_plus

from changelog.exceptions import ChangelogError, ChangelogMissingConfigError, ChangelogValidationError
from changelog.utils import reverse_format

ChangeType = Literal["Security", "Deprecated", "Added", "Changed", "Removed", "Fixed"]


class Bump(Enum):
    MAJOR = 0
    MINOR = 1
    PATCH = 2


class ReleaseTag(str):
    @classmethod
    def from_semver(cls, semver: tuple[int, int, int]) -> ReleaseTag:
        return cls(".".join(map(str, semver)))

    @property
    def is_semver(self):
        return re.match(r"^\d+\.\d+\.\d+$", self)

    @property
    def semver(self) -> tuple[int, int, int]:
        if not self.is_semver:
            raise TypeError(f"ReleaseTag {self!r} is not a semantic version")
        return cast(tuple[int, int, int], tuple(map(int, self.split("."))))

    def bump_semver(self, bump: Bump) -> ReleaseTag:
        semver = list(self.semver)
        semver[bump.value] = semver[bump.value] + 1
        for idx in range(bump.value + 1, 3):
            semver[idx] = 0
        return type(self)(".".join(map(str, semver)))


_UNRELEASED = ReleaseTag("Unreleased")


@dataclass
class ChangelogConfig:
    release_link_format: Optional[str] = None
    breaking_change_token: str = field(default="BREAKING", metadata={"parse": unquote_plus, "render": quote_plus})

    @property
    def fields(self) -> dict[str, Field]:
        return self.__dataclass_fields__  # type: ignore[attr-defined]

    def get(self, key: str, default: str = None) -> str:
        if key not in self.fields:
            raise ChangelogError(f"{key} is not a valid config option.")
        value = getattr(self, key)
        if not value:
            if default:
                return default
            raise ChangelogMissingConfigError(f"No config option for {key!r}", field=key)
        return value

    def set(self, key: str, value: str) -> None:
        if key not in self.fields:
            raise ChangelogError(f"{key} is not a valid config option.")
        setattr(self, key, value)


@dataclass
class Changelog:
    header: str = ""
    config: ChangelogConfig = field(default_factory=ChangelogConfig)
    releases: OrderedDict[ReleaseTag, ReleaseSection] = field(default_factory=OrderedDict)
    links: OrderedDict[str, str] = field(default_factory=OrderedDict)

    def validate(self):
        """Validate the changelog."""
        if not self.releases:
            raise ChangelogValidationError("Changelog contains no releases!")
        missing_tag_links = set(self.releases) - set(self.links)
        if missing_tag_links:
            raise ChangelogValidationError(f"The following releases are missing links: {missing_tag_links}")

    def add_entry(self, change_type: ChangeType, *items: str, breaking: bool = False, tag: str = None) -> None:
        """Add an entry to the changelog, under unreleased."""
        tag = ReleaseTag(tag) if tag else _UNRELEASED
        assert change_type in ChangeType.__args__  # type: ignore
        prefix = f"{self.config.get('breaking_change_token')} " if breaking else ""
        self.releases.setdefault(tag, ReleaseSection(entries={}, timestamp=None)).entries.setdefault(
            change_type, []
        ).append(Entry(text=prefix + items[0], children=[Entry(text=item) for item in items[1:]]))

    @property
    def latest_tag(self) -> Optional[ReleaseTag]:
        return next((tag for tag in self.releases if not tag == _UNRELEASED), None)

    def next_tag(self, force: Bump = None) -> ReleaseTag:
        if not self.latest_tag:
            return ReleaseTag.from_semver((0, 1, 0))
        if not self.latest_tag.is_semver:
            raise ChangelogError(f"Previous tag {self.latest_tag} is not semantic")
        if force:
            return self.latest_tag.bump_semver(force)
        unreleased = self.releases[_UNRELEASED].entries
        if self.config.get("breaking_change_token", "BREAKING") in str(unreleased) and self.latest_tag.semver[0] > 0:
            return self.latest_tag.bump_semver(Bump.MAJOR)
        if set(unreleased) - {"Fixed"}:
            return self.latest_tag.bump_semver(Bump.MINOR)
        return self.latest_tag.bump_semver(Bump.PATCH)

    def cut_release(self, force: Bump = None, tag: str = None) -> tuple[ReleaseTag, ReleaseSection]:
        previous_tag = self.latest_tag
        release_tag = ReleaseTag(tag) if tag else self.next_tag(force=force)
        # Move entries from unreleased to the new tag:
        self.releases[release_tag] = self.releases[_UNRELEASED]
        self.releases[release_tag].timestamp = date.today().isoformat()
        self.releases[_UNRELEASED] = ReleaseSection(entries={}, timestamp=None)
        # Reorder releases:
        self.releases.move_to_end(release_tag, last=False)
        self.releases.move_to_end(_UNRELEASED, last=False)
        # Update tag links
        link_spec = self.config.get("release_link_format")
        self.links[release_tag] = link_spec.format(previous_tag=previous_tag or "initial", tag=release_tag)
        self.links[_UNRELEASED] = link_spec.format(
            previous_tag=release_tag,
            tag=reverse_format(self.links[_UNRELEASED], link_spec, cast(dict[str, str], {})).get("tag", "HEAD"),
        )
        # Reorder links
        self.links.move_to_end(release_tag, last=False)
        self.links.move_to_end(_UNRELEASED, last=False)
        return release_tag, self.releases[release_tag]


@dataclass
class ReleaseSection:
    entries: dict[str, list[Entry]]
    timestamp: Optional[str]


@dataclass
class Entry:
    text: str
    children: list[Entry] = field(default_factory=list)



from collections import OrderedDict

from changelog.model import Changelog, ChangelogConfig, ReleaseSection, ReleaseTag


def default_changelog(release_link_format: str, breaking_change_token: str = "BREAKING"):
    """Generate an initial changelog with default header and links."""
    return Changelog(
        header="""# Changelog
All notable changes to this project will be documented in this file.

The format is based on [Keep a Changelog] and this project adheres to
[Semantic Versioning].

Types of changes are:
* **Security** in case of vulnerabilities.
* **Deprecated** for soon-to-be removed features.
* **Added** for new features.
* **Changed** for changes in existing functionality.
* **Removed** for now removed features.
* **Fixed** for any bug fixes.""",
        config=ChangelogConfig(
            release_link_format=release_link_format,
            breaking_change_token=breaking_change_token,
        ),
        releases=OrderedDict(
            {
                ReleaseTag("Unreleased"): ReleaseSection(entries={}, timestamp=None),
            }
        ),
        links=OrderedDict(
            {
                "Unreleased": release_link_format.format(previous_tag="initial", tag="HEAD"),
                "Keep a Changelog": "http://keepachangelog.com/en/1.0.0/",
                "Semantic Versioning": "http://semver.org/spec/v2.0.0.html",
            },
        ),
    )



from __future__ import annotations

from dataclasses import asdict

from changelog.model import Changelog, ChangelogConfig, Entry, ReleaseSection, ReleaseTag


def dumps(changelog: Changelog, indent: int = 2) -> str:
    changelog.validate()
    return (
        "\n\n".join(
            [
                changelog.header.strip(),
                _render_changelog_releases(changelog.releases, indent=indent),
                _render_changelog_links(changelog.links, set(changelog.releases)),
                _render_changelog_config(changelog.config),
            ]
        )
        + "\n"
    )


def _render_changelog_releases(releases: dict[ReleaseTag, ReleaseSection], indent: int = 2) -> str:
    return "\n\n".join([render_changelog_release(release_tag, section) for release_tag, section in releases.items()])


def render_changelog_release(release_tag: ReleaseTag, section: ReleaseSection, indent: int = 2) -> str:
    header = f"## [{release_tag}]"
    if section.timestamp:
        header += f" - {section.timestamp}"
    return "\n".join([header, _render_changelog_change_types(section.entries, indent=indent)])


def _render_changelog_change_types(change_types: dict[str, list[Entry]], indent: int = 2) -> str:
    return "\n\n".join(
        [
            _render_changelog_change_type(change_type, entries, indent=indent)
            for change_type, entries in sorted(change_types.items())
        ]
    )


def _render_changelog_change_type(change_type: str, entries: list[Entry], indent: int = 2) -> str:
    return "\n".join(
        [f"### {change_type}", "\n".join([_render_changelog_entry(entry, indent=indent) for entry in entries])]
    )


def _render_changelog_entry(entry: Entry, indent: int = 2, _indent_level: int = 0) -> str:
    bullet = ("*", "-", "+")[_indent_level % 3]
    return "\n".join(
        [
            " " * indent * _indent_level + f"{bullet} {entry.text}",
            *[
                _render_changelog_entry(sub_entry, indent=indent, _indent_level=_indent_level + 1)
                for sub_entry in entry.children
            ],
        ]
    )


def _render_changelog_links(links: dict[str, str], release_tags: set[ReleaseTag]) -> str:
    return "\n\n".join(
        [
            "\n".join(
                [
                    f"[{link_name}]: {link_target}"
                    for link_name, link_target in links.items()
                    if link_name in release_tags
                ]
            ),
            "\n".join(
                [
                    f"[{link_name}]: {link_target}"
                    for link_name, link_target in links.items()
                    if link_name not in release_tags
                ]
            ),
        ]
    )


def _render_changelog_config(config: ChangelogConfig) -> str:
    return "\n".join([_render_config_field(config, field, value) for field, value in asdict(config).items() if value])


def _render_config_field(config: ChangelogConfig, field: str, value: str) -> str:
    render_value = config.fields[field].metadata.get("render", lambda _: _)
    return f"[_{field}]: {render_value(value)}"


def dump_to_file(changelog: Changelog, path: str = "CHANGELOG.md") -> None:
    with open(path, "w") as file:
        file.write(dumps(changelog))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['changelog', 'changelog.cli']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['changelog = changelog.__main__:app']}

setup_kwargs = {
    'name': 'changelog-cmd',
    'version': '0.1.1',
    'description': '',
    'long_description': '# changelog-cmd\n\n![example workflow](https://github.com/jacksmith15/changelog-cmd/actions/workflows/main.yml/badge.svg)\n\nTool for managing a changelog in the style of [Keep a Changelog].\n\n### Features:\n\n- Write new changelog entries in a consistent manner\n- Validate and format your changelog\n- Suggest semantic version of releases based on their contents\n- Write scripts which make automated changes and updates to changelogs, whilst keeping the changelog in a human-readable format\n\n## Installation\n\nTODO\n\n## Usage\n\n### Starting a changelog\n\nCreate a changelog named `CHANGELOG.md` in the current directory by running\n\n```shell\nchangelog init --release-link-format=RELEASE_LINK_FORMAT\n```\n\nWhere the `--release-link-format` option specifies how to generate links to a given release, and is given using a Python string format specifier. For example, if you want to link to source on GitHub, this format specifier might be:\n\n- `https://github.com/user/repo/tree/{tag}` - the source code at this release\n- `https://github.com/user/repo/compare/{previous_tag}...{tag}` - a comparison of this release with the one before\n\nOr if you want to link to PyPi, the format might be:\n\n- `https://pypi.org/project/package/{tag}/`\n\nAbove you see the two variables available for substitution:\n- `tag` is the tag of the release\n- `previous_tag` is the tag of the previous release\n\n\n### Finding the changelog\n\nBy default, `changelog` looks for a changelog called `CHANGELOG.md` in the current directory. You can override this by passing a command-line option:\n\n```shell\nchangelog --path path/to/changelog.md ...\n```\n\nOr you can set an environment variable:\n\n```shell\nCHANGELOG_PATH=path/to/changelog.md ...\n```\n\nThe command-line option takes precedence.\n\n### Adding entries to a changelog\n\nAdd an entry to the unreleased section of a changelog:\n\n```shell\nchangelog entry added --message "Description of my change"\n```\n\nYou can replace `added` in the above with any of the following types of entry:\n\n- `added`\n- `changed`\n- `deprecated`\n- `fixed`\n- `removed`\n- `security`\n\nTo specify a change as breaking, simply include the `--breaking` flag:\n\n```shell\nchangelog entry changed --message "Description of a breaking change" --breaking\n```\n\nIf you need to add a missing entry to a past release, you can specify the release tag explicitly:\n\n```shell\nchangelog entry fixed --message "Description of a fix" --tag "0.1.1"\n```\n\n### Cutting a release\n\nWhen you are ready to cut a release, run the following:\n\n```shell\nchangelog release\n```\n\nThis will identify the correct semantic version. Alternatively, you can specify the semantic version bump yourself:\n\n```shell\nchangelog release --bump "major"\n```\n\nOr, if you don\'t use semantic versioning at all, simply specify the tag of the release:\n\n```shell\nchangelog release --tag "2021.r3"\n```\n\n### Formatting and validation\n\nUsing this tool should not preclude manual editing of a changelog. To ensure that manual changes don\'t break conventions, you can use the following two commands in your development workflow:\n\nSimply check that the changelog is still in a recognisable format:\n```shell\nchangelog validate\n```\n\nFormat the changelog in a standard manner:\n```shell\nchangelog format\n```\n\n### Changelog configuration\n\nThis tool stores configuration in the changelog itself. The currently available config fields are:\n\n- `release_link_format` specifies the format for release links\n- `breaking_change_token` specifies the token for breaking changes\n\nCheck the value of a config field via:\n\n```shell\nchangelog config get --field CONFIG_FIELD\n```\n\nAnd set it via\n\n```shell\nchangelog config set --field CONFIG_FIELD --value VALUE\n```\n\n## Development\n\nInstall dependencies:\n\n```shell\npyenv shell 3.9.4  # Or other 3.9.x\npre-commit install  # Configure commit hooks\npoetry install  # Install Python dependencies\n```\n\nRun tests:\n\n```shell\npoetry run inv verify\n```\n\n## Future improvements\n\nThe following is a list of possible future improvements for this tool:\n\n- Support a release-per-change workflow (every change is tagged as a new release)\n- Support configuration of of change types beyond those specified by [Keep a Changelog]\n- Extensions to help interaction with `git`, e.g. merge conflict resolution and/or validation.\n\n## License\nThis project is distributed under the MIT license.\n\n[Keep a Changelog]: http://keepachangelog.com/en/1.0.0/',
    'author': 'Jack Smith',
    'author_email': 'jack.je.smith@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jacksmith15/changelog-cmd',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
